/**
 * This file contains the functions associated with the logging and parameter tables
 * of contents for the CCrazyflie class.
 *
 */
#include "CCrazyflie.h"
#include "CTOC.h"

#include <time.h>
#include <iostream>
#include <stdio.h>
#include <errno.h>
#include <string.h>

// Various headers for the files
static const char *variables = "%Time\t\tMarker\t\tMotor_1\t\t"
							   "Motor_2\t\tMotor_3\t\t"
							   "Motor_4\t\tPitch\t\t"
							   "Pitch_err\t\tPitch_pid_p\t\t"
							   "Pitch_pid_d\t\t"
							   "Roll\t\tRoll_err\t\t"
							   "Roll_pid_p\t\tRoll_pid_d\t\t";

static const char *units = "&sec\t\tMarker\t\t%thrust\t\t"
						   "%thrust\t\t%thrust\t\t"
						   "%thrust\t\tdegrees\t\t"
						   "degrees\t\tdegrees\t\t"
						   "degrees\t\t"
						   "degrees\t\tdegrees\t\t"
						   "degrees\t\tdegrees\t\t";

/*
 * Open a log file for the quadcopter
 *
 * @param baseFileName The base filename for the logfile
 */
void CCrazyflie::openLogFile(char *baseFileName) {
	char logFileName[40];

	char timeString[40];
	time_t rawtime;
	time( &rawtime );
	struct tm *timeinfo = localtime( &rawtime );
	strftime(timeString, 40, "%Y_%m_%d_%H:%M:%S", timeinfo);

	// Create the log file name (including date)
	sprintf(logFileName, "%s_%s.txt", baseFileName, timeString);

	printf("Opening log file %s  for quadcopter %d...", logFileName, (m_quadNum + 1));

	try {
		file_log.open(logFileName, std::ios_base::out);
	} catch (...) {
		std::cout << "Error opening logfile for quadcopter " << m_quadNum;
		exit(-1);
	}

	// Place the header information into the log file
	file_log << "#Crazyflie" << std::endl;

	// Add the controller type to the logfile
	file_log << "#" << this->getControllerTypeString() << std::endl;
	file_log << variables << std::endl;
	file_log << units << std::endl;

	std::cout << " Complete" << std::endl;
}

/*
 * Write data to the log file
 */
void CCrazyflie::writeLogData() {
	// Make a new line
	file_log << std::endl;

	// Print the time
	file_log << this->currentTime();

	// Print the marker and motors
	file_log << "\t\t" << 0;

	file_log << "\t\t" << this->sensorDoubleValue("pwm.m1_pwm");
	file_log << "\t\t" << this->sensorDoubleValue("pwm.m2_pwm");
	file_log << "\t\t" << this->sensorDoubleValue("pwm.m3_pwm");
	file_log << "\t\t" << this->sensorDoubleValue("pwm.m4_pwm");

	// Print pitch and roll data
	float pitch = this->getQuadPitch();
	file_log << "\t\t" << pitch;
	file_log << "\t\t" << -pitch;

	file_log << "\t\t" << this->sensorDoubleValue("pid_attitude.pitch_outP");
	file_log << "\t\t" << this->sensorDoubleValue("pid_attitude.pitch_outD");

	float roll = this->getQuadRoll();
	file_log << "\t\t" << roll;
	file_log << "\t\t" << -roll;

	file_log << "\t\t" << this->sensorDoubleValue("pid_attitude.roll_outP");
	file_log << "\t\t" << this->sensorDoubleValue("pid_attitude.roll_outD");
	file_log << "\t\t" << this->sensorDoubleValue("ctrlStdnt.rollRate");
}


bool CCrazyflie::readTOCParameters() {
	if( ALL_THE_DEBUG ) printf( "Enter: %s\n", __FUNCTION__  );
	printf( "      readTocParameters requestMetaData\n" );
	if (m_tocParameters->requestMetaData()) {
		printf( "      requestItems\n" );
		if (m_tocParameters->requestItems()) {
			printf( "      return true\n" );
			return true;
		}
	}

	if( ALL_THE_DEBUG ) printf( "Exit: %s\n", __FUNCTION__  );
	return false;
}

bool CCrazyflie::readTOCLogs() {
	if( ALL_THE_DEBUG ) printf( "Enter: %s\n", __FUNCTION__  );
	if (m_tocLogs->requestMetaData()) {
		if (m_tocLogs->requestItems()) {
			return true;
		}
	}

	if( ALL_THE_DEBUG ) printf( "Exit: %s\n", __FUNCTION__  );
	return false;
}

void CCrazyflie::displayLoggingBlocksInitialized() {
	if( ALL_THE_DEBUG ) printf( "Enter: %s\n", __FUNCTION__  );
	m_tocLogs->printLoggingBlocksInitialized();
	if( ALL_THE_DEBUG ) printf( "Exit: %s\n", __FUNCTION__  );
}

double CCrazyflie::sensorDoubleValue(std::string strName) {
	return m_tocLogs->doubleValue(strName);
}

bool CCrazyflie::addLoggingBlock(const char *name, uint16_t frequency) {
	bool retval = m_tocLogs->registerLoggingBlock(name, frequency);
	return(retval);
}

void CCrazyflie::removeLoggingBlock(const char *name) {
	if( ALL_THE_DEBUG ) printf( "Enter: %s\n", __FUNCTION__  );
	m_tocLogs->unregisterLoggingBlock(name);

	if( ALL_THE_DEBUG ) printf( "Exit: %s\n", __FUNCTION__  );
}

bool CCrazyflie::addLoggingEntry(const char *blockName, const char *entryName) {
	if( ALL_THE_DEBUG ) printf( "Enter: %s\n", __FUNCTION__  );
	bool retval = m_tocLogs->startLogging(entryName, blockName);

	if (!retval) {
		std::cout << "Error starting logging entry " << entryName << " in block " << blockName << std::endl;
	}

	if( ALL_THE_DEBUG ) printf( "Exit: %s\n", __FUNCTION__  );
	return ( retval );
}